/*
 * livepatch_bsc1099258
 *
 * Fix for CVE-2018-12904, bsc#1099258
 *
 *  Upstream commit:
 *  727ba748e110 ("kvm: nVMX: Enforce cpl=0 for VMX instructions")
 *
 *  SLE12 commit:
 *  not affected
 *
 *  SLE12-SP1 commit
 *  not affected
 *
 *  SLE12-SP2 commit:
 *  not affected
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE15 commit:
 *  a1ac0dc3929e729729a725efb4236d4d01aec294
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */


#if IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/kvm_host.h>
#include <asm/vmx.h>
#include <linux/highmem.h>
#include <asm/segment.h>
#include <asm/nospec-branch.h>
#include <linux/nospec.h>
#include "livepatch_bsc1099258.h"

#if !IS_MODULE(CONFIG_KVM)
#error "Live patch supports only CONFIG_KVM=m"
#endif

#define KLP_PATCHED_MODULE "kvm_intel"


struct vcpu_vmx;
struct vmcs12;
struct loaded_vmcs;

struct vmx_capability {
	u32 ept;
	u32 vpid;
};

static struct kvm_x86_ops **klp_kvm_x86_ops;
static bool *klp_kvm_rebooting;
static asmlinkage void (*klp_kvm_spurious_fault)(void);
static void (*klp_kvm_queue_exception)(struct kvm_vcpu *vcpu, unsigned nr);
static struct page *(*klp_kvm_vcpu_gfn_to_page)(struct kvm_vcpu *vcpu,
						gfn_t gfn);
static void (*klp_kvm_release_page_clean)(struct page *page);
static int (*klp_kvm_vcpu_write_guest_page)(struct kvm_vcpu *vcpu, gfn_t gfn,
					    const void *data, int offset,
					    int len);
static int (*klp_kvm_vcpu_write_guest)(struct kvm_vcpu *vcpu, gpa_t gpa,
				       const void *data, unsigned long len);
static int (*klp_kvm_write_guest_virt_system)(struct x86_emulate_ctxt *ctxt,
					      gva_t addr, void *val,
					      unsigned int bytes,
					      struct x86_exception *exception);
static int (*klp_kvm_read_guest_virt)(struct x86_emulate_ctxt *ctxt,
				      gva_t addr, void *val, unsigned int bytes,
				      struct x86_exception *exception);
static void (*klp_kvm_inject_page_fault)(struct kvm_vcpu *vcpu,
					 struct x86_exception *fault);
static void (*klp_kvm_mmu_sync_roots)(struct kvm_vcpu *vcpu);
static void (*klp_kvm_queue_exception_e)(struct kvm_vcpu *vcpu, unsigned nr,
					 u32 error_code);

static bool *klp_enable_shadow_vmcs;
static bool *klp_enable_ept;
static unsigned short (*klp_vmcs_field_to_offset_table)[HOST_RIP + 1];
static struct vmx_capability *klp_vmx_capability;
static void (*klp_vmwrite_error)(unsigned long field, unsigned long value);
static void (*klp_nested_vmx_failInvalid)(struct kvm_vcpu *vcpu);
static void (*klp_copy_shadow_to_vmcs12)(struct vcpu_vmx *vmx);
static void (*klp_nested_vmx_failValid)(struct kvm_vcpu *vcpu,
					u32 vm_instruction_error);
static void (*klp_nested_vmx_succeed)(struct kvm_vcpu *vcpu);
static int (*klp_kvm_skip_emulated_instruction)(struct kvm_vcpu *vcpu);
static int (*klp_nested_vmx_get_vmptr)(struct kvm_vcpu *vcpu, gpa_t *vmpointer);
static int (*klp_alloc_loaded_vmcs)(struct loaded_vmcs *loaded_vmcs);
static struct vmcs *(*klp_alloc_vmcs_cpu)(int cpu);
static void (*klp_vmcs_clear)(struct vmcs *vmcs);
static void (*klp_free_loaded_vmcs)(struct loaded_vmcs *loaded_vmcs);
static void (*klp_free_nested)(struct vcpu_vmx *vmx);
static int (*klp_get_vmx_mem_address)(struct kvm_vcpu *vcpu,
				      unsigned long exit_qualification,
				      u32 vmx_instruction_info, bool wr,
				      gva_t *ret);
static u64 (*klp_construct_eptp)(struct kvm_vcpu *vcpu, unsigned long root_hpa);
static enum hrtimer_restart
(*klp_vmx_preemption_timer_fn)(struct hrtimer *timer);
static int (*klp_vmx_get_cpl)(struct kvm_vcpu *vcpu);

static struct {
	char *name;
	void **addr;
} klp_funcs[] = {
	{ "kvm:kvm_x86_ops", (void *)&klp_kvm_x86_ops },
	{ "kvm:kvm_rebooting", (void *)&klp_kvm_rebooting },
	{ "kvm:kvm_spurious_fault", (void *)&klp_kvm_spurious_fault },
	{ "kvm:kvm_queue_exception", (void *)&klp_kvm_queue_exception },
	{ "kvm:kvm_skip_emulated_instruction",
			(void *)&klp_kvm_skip_emulated_instruction },
	{ "kvm:kvm_vcpu_gfn_to_page",
			(void *)&klp_kvm_vcpu_gfn_to_page },
	{ "kvm:kvm_release_page_clean",
			(void *)&klp_kvm_release_page_clean },
	{ "kvm:kvm_vcpu_write_guest_page",
			(void *)&klp_kvm_vcpu_write_guest_page },
	{ "kvm:kvm_vcpu_write_guest", (void *)&klp_kvm_vcpu_write_guest },
	{ "kvm:kvm_write_guest_virt_system",
			(void *)&klp_kvm_write_guest_virt_system },
	{ "kvm:kvm_read_guest_virt", (void *)&klp_kvm_read_guest_virt },
	{ "kvm:kvm_inject_page_fault", (void *)&klp_kvm_inject_page_fault },
	{ "kvm:kvm_mmu_sync_roots", (void *)&klp_kvm_mmu_sync_roots },
	{ "kvm:kvm_queue_exception_e", (void *)&klp_kvm_queue_exception_e },

	{ "kvm_intel:enable_shadow_vmcs", (void *)&klp_enable_shadow_vmcs },
	{ "kvm_intel:enable_ept", (void *)&klp_enable_ept },
	{ "kvm_intel:vmcs_field_to_offset_table",
			(void *)&klp_vmcs_field_to_offset_table },
	{ "kvm_intel:vmx_capability", (void *)&klp_vmx_capability },
	{ "kvm_intel:vmwrite_error", (void *)&klp_vmwrite_error },
	{ "kvm_intel:nested_vmx_failInvalid",
			(void *)&klp_nested_vmx_failInvalid },
	{ "kvm_intel:copy_shadow_to_vmcs12",
			(void *)&klp_copy_shadow_to_vmcs12 },
	{ "kvm_intel:nested_vmx_failValid",
			(void *)&klp_nested_vmx_failValid },
	{ "kvm_intel:nested_vmx_succeed", (void *)&klp_nested_vmx_succeed },
	{ "kvm_intel:nested_vmx_get_vmptr",
			(void *)&klp_nested_vmx_get_vmptr },
	{ "kvm_intel:alloc_loaded_vmcs", (void *)&klp_alloc_loaded_vmcs },
	{ "kvm_intel:alloc_vmcs_cpu", (void *)&klp_alloc_vmcs_cpu },
	{ "kvm_intel:vmcs_clear", (void *)&klp_vmcs_clear },
	{ "kvm_intel:free_loaded_vmcs", (void *)&klp_free_loaded_vmcs },
	{ "kvm_intel:free_nested", (void *)&klp_free_nested },
	{ "kvm_intel:get_vmx_mem_address", (void *)&klp_get_vmx_mem_address },
	{ "kvm_intel:construct_eptp", (void *)&klp_construct_eptp },
	{ "kvm_intel:vmx_preemption_timer_fn",
			(void *)&klp_vmx_preemption_timer_fn },
	{ "kvm_intel:vmx_get_cpl", (void *)&klp_vmx_get_cpl },
};


/* from arch/x86/include/asm/kvm_host.h */
/*
 * Avoid module dependency on kvm.ko.
 *
 * This is the original macro but with the load from kvm_rebooting
 * replaced by an indirect load from *klp_kvm_rebooting and the call
 * to kvm_spurious_fault() replaced by an indirect call to
 * *klp_kvm_spurious_fault.
 */
#define klp____kvm_handle_fault_on_reboot(insn, cleanup_insn)	\
	"666: " insn "\n\t"				\
	"668: \n\t"					\
	".pushsection .fixup, \"ax\" \n"		\
	"667: \n\t"					\
	cleanup_insn "\n\t"				\
	"pushq %%rax \n\t"				\
	"movq klp_kvm_rebooting, %%rax\n\t"		\
	"cmpb $0, (%%rax) \n\t"			\
	"popq %%rax \n\t"				\
	"jne 668b \n\t"				\
	__ASM_SIZE(push) " $666b \n\t"			\
	/* this depends on [thunk_target] "a" constraints */	\
	"movq klp_kvm_spurious_fault, %%rax \n\t"	\
	CALL_NOSPEC "\n\t"				\
	/* kvm_spurious_fault() is noreturn. */	\
	/* Silence objtool by balancing the stack */	\
	/* and jmp'ing somewhere */			\
	"popq %%rax\n\t"				\
	"jmp 668b \n\t"				\
	".popsection \n\t"				\
	_ASM_EXTABLE(666b, 667b)

#define klp__kvm_handle_fault_on_reboot(insn)		\
	klp____kvm_handle_fault_on_reboot(insn, "")

/* Avoid module dependency on kvm.ko. */
static inline struct page *klp_kvm_vcpu_gpa_to_page(struct kvm_vcpu *vcpu,
						    gpa_t gpa)
{
	return klp_kvm_vcpu_gfn_to_page(vcpu, gpa_to_gfn(gpa));
}

/* Avoid module dependency on kvm.ko. */
static inline void klp_kvm_inject_gp(struct kvm_vcpu *vcpu, u32 error_code)
{
	klp_kvm_queue_exception_e(vcpu, GP_VECTOR, error_code);
}


/* from arch/x86/kvm/kvm_cache_regs.h */
#define KLP_KVM_POSSIBLE_CR4_GUEST_BITS				  \
	(X86_CR4_PVI | X86_CR4_DE | X86_CR4_PCE | X86_CR4_OSFXSR  \
	 | X86_CR4_OSXMMEXCPT | X86_CR4_LA57 | X86_CR4_PGE)

/* inlined */
static inline unsigned long klp_kvm_register_read(struct kvm_vcpu *vcpu,
						  enum kvm_reg reg)
{
	if (!test_bit(reg, (unsigned long *)&vcpu->arch.regs_avail))
		(*klp_kvm_x86_ops)->cache_reg(vcpu, reg);

	return vcpu->arch.regs[reg];
}

/* inlined */
static inline void klp_kvm_register_write(struct kvm_vcpu *vcpu,
					  enum kvm_reg reg,
					  unsigned long val)
{
	vcpu->arch.regs[reg] = val;
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_dirty);
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_avail);
}

/* inlined */
static inline ulong klp_kvm_read_cr4_bits(struct kvm_vcpu *vcpu, ulong mask)
{
	ulong tmask = mask & KLP_KVM_POSSIBLE_CR4_GUEST_BITS;
	if (tmask & vcpu->arch.cr4_guest_owned_bits)
		(*klp_kvm_x86_ops)->decache_cr4_guest_bits(vcpu);
	return vcpu->arch.cr4 & mask;
}


/* from arch/x86/kvm/x86.h */
/* inlined */
static inline int klp_is_long_mode(struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return vcpu->arch.efer & EFER_LMA;
#else
	return 0;
#endif
}

/* inlined */
static bool klp_is_64_bit_mode(struct kvm_vcpu *vcpu)
{
	int cs_db, cs_l;

	if (!klp_is_long_mode(vcpu))
		return false;
	(*klp_kvm_x86_ops)->get_cs_db_l_bits(vcpu, &cs_db, &cs_l);
	return cs_l;
}

/* inlined */
static inline u8 klp_vcpu_virt_addr_bits(struct kvm_vcpu *vcpu)
{
	return klp_kvm_read_cr4_bits(vcpu, X86_CR4_LA57) ? 57 : 48;
}

/* inlined */
static inline u64 klp_get_canonical(u64 la, u8 vaddr_bits)
{
	return ((int64_t)la << (64 - vaddr_bits)) >> (64 - vaddr_bits);
}

/* inlined */
static inline bool klp_is_noncanonical_address(u64 la, struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return klp_get_canonical(la, klp_vcpu_virt_addr_bits(vcpu)) != la;
#else
	return false;
#endif
}

/* inlined */
static inline unsigned long klp_kvm_register_readl(struct kvm_vcpu *vcpu,
						   enum kvm_reg reg)
{
	unsigned long val = klp_kvm_register_read(vcpu, reg);

	return klp_is_64_bit_mode(vcpu) ? val : (u32)val;
}

/* inlined */
static inline void klp_kvm_register_writel(struct kvm_vcpu *vcpu,
					   enum kvm_reg reg,
					   unsigned long val)
{
	if (!klp_is_64_bit_mode(vcpu))
		val = (u32)val;
	return klp_kvm_register_write(vcpu, reg, val);
}


/* from arch/x86/kvm/cpuid.h */
/* inlined */
static inline int klp_cpuid_maxphyaddr(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.maxphyaddr;
}


/* from arch/x86/kvm/vmx.c */
#define klp__ex(x) klp__kvm_handle_fault_on_reboot(x)
#define klp__ex_clear(x, reg) \
	klp____kvm_handle_fault_on_reboot(x, "xor " reg " , " reg)

#define KLP_VMX_VPID_EXTENT_SUPPORTED_MASK		\
	(VMX_VPID_EXTENT_INDIVIDUAL_ADDR_BIT |	\
	VMX_VPID_EXTENT_SINGLE_CONTEXT_BIT |	\
	VMX_VPID_EXTENT_GLOBAL_CONTEXT_BIT |	\
	VMX_VPID_EXTENT_SINGLE_NON_GLOBAL_BIT)

#define KLP_NR_AUTOLOAD_MSRS 8

struct vmcs {
	u32 revision_id;
	u32 abort;
	char data[0];
};

struct loaded_vmcs {
	struct vmcs *vmcs;
	struct vmcs *shadow_vmcs;
	int cpu;
	bool launched;
	bool nmi_known_unmasked;
	unsigned long vmcs_host_cr3;	/* May not match real cr3 */
	unsigned long vmcs_host_cr4;	/* May not match real cr4 */
	/* Support for vnmi-less CPUs */
	int soft_vnmi_blocked;
	ktime_t entry_time;
	s64 vnmi_blocked_time;
	unsigned long *msr_bitmap;
	struct list_head loaded_vmcss_on_cpu_link;
};

#define KLP_VMCS12_REVISION 0x11e57ed0

#define KLP_VMCS12_SIZE 0x1000

struct nested_vmx {
	/* Has the level1 guest done vmxon? */
	bool vmxon;
	gpa_t vmxon_ptr;
	bool pml_full;

	/* The guest-physical address of the current VMCS L1 keeps for L2 */
	gpa_t current_vmptr;
	/*
	 * Cache of the guest's VMCS, existing outside of guest memory.
	 * Loaded from guest memory during VMPTRLD. Flushed to guest
	 * memory during VMCLEAR and VMPTRLD.
	 */
	struct vmcs12 *cached_vmcs12;
	/*
	 * Indicates if the shadow vmcs must be updated with the
	 * data hold by vmcs12
	 */
	bool sync_shadow_vmcs;

	bool change_vmcs01_virtual_x2apic_mode;
	/* L2 must run next, and mustn't decide to exit to L1. */
	bool nested_run_pending;

	struct loaded_vmcs vmcs02;

	/*
	 * Guest pages referred to in the vmcs02 with host-physical
	 * pointers, so we must keep them pinned while L2 runs.
	 */
	struct page *apic_access_page;
	struct page *virtual_apic_page;
	struct page *pi_desc_page;
	struct pi_desc *pi_desc;
	bool pi_pending;
	u16 posted_intr_nv;

	struct hrtimer preemption_timer;
	bool preemption_timer_expired;

	/* to migrate it to L2 if VM_ENTRY_LOAD_DEBUG_CONTROLS is off */
	u64 vmcs01_debugctl;

	u16 vpid02;
	u16 last_vpid;

	/*
	 * We only store the "true" versions of the VMX capability MSRs. We
	 * generate the "non-true" versions by setting the must-be-1 bits
	 * according to the SDM.
	 */
	u32 nested_vmx_procbased_ctls_low;
	u32 nested_vmx_procbased_ctls_high;
	u32 nested_vmx_secondary_ctls_low;
	u32 nested_vmx_secondary_ctls_high;
	u32 nested_vmx_pinbased_ctls_low;
	u32 nested_vmx_pinbased_ctls_high;
	u32 nested_vmx_exit_ctls_low;
	u32 nested_vmx_exit_ctls_high;
	u32 nested_vmx_entry_ctls_low;
	u32 nested_vmx_entry_ctls_high;
	u32 nested_vmx_misc_low;
	u32 nested_vmx_misc_high;
	u32 nested_vmx_ept_caps;
	u32 nested_vmx_vpid_caps;
	u64 nested_vmx_basic;
	u64 nested_vmx_cr0_fixed0;
	u64 nested_vmx_cr0_fixed1;
	u64 nested_vmx_cr4_fixed0;
	u64 nested_vmx_cr4_fixed1;
	u64 nested_vmx_vmcs_enum;
	u64 nested_vmx_vmfunc_controls;

	/* SMM related state */
	struct {
		/* in VMX operation on SMM entry? */
		bool vmxon;
		/* in guest mode on SMM entry? */
		bool guest_mode;
	} smm;
};

typedef u64 natural_width;
struct __packed vmcs12 {
	/* According to the Intel spec, a VMCS region must start with the
	 * following two fields. Then follow implementation-specific data.
	 */
	u32 revision_id;
	u32 abort;

	u32 launch_state; /* set to 0 by VMCLEAR, to 1 by VMLAUNCH */
	u32 padding[7]; /* room for future expansion */

	u64 io_bitmap_a;
	u64 io_bitmap_b;
	u64 msr_bitmap;
	u64 vm_exit_msr_store_addr;
	u64 vm_exit_msr_load_addr;
	u64 vm_entry_msr_load_addr;
	u64 tsc_offset;
	u64 virtual_apic_page_addr;
	u64 apic_access_addr;
	u64 posted_intr_desc_addr;
	u64 vm_function_control;
	u64 ept_pointer;
	u64 eoi_exit_bitmap0;
	u64 eoi_exit_bitmap1;
	u64 eoi_exit_bitmap2;
	u64 eoi_exit_bitmap3;
	u64 eptp_list_address;
	u64 xss_exit_bitmap;
	u64 guest_physical_address;
	u64 vmcs_link_pointer;
	u64 pml_address;
	u64 guest_ia32_debugctl;
	u64 guest_ia32_pat;
	u64 guest_ia32_efer;
	u64 guest_ia32_perf_global_ctrl;
	u64 guest_pdptr0;
	u64 guest_pdptr1;
	u64 guest_pdptr2;
	u64 guest_pdptr3;
	u64 guest_bndcfgs;
	u64 host_ia32_pat;
	u64 host_ia32_efer;
	u64 host_ia32_perf_global_ctrl;
	u64 padding64[8]; /* room for future expansion */
	/*
	 * To allow migration of L1 (complete with its L2 guests) between
	 * machines of different natural widths (32 or 64 bit), we cannot have
	 * unsigned long fields with no explict size. We use u64 (aliased
	 * natural_width) instead. Luckily, x86 is little-endian.
	 */
	natural_width cr0_guest_host_mask;
	natural_width cr4_guest_host_mask;
	natural_width cr0_read_shadow;
	natural_width cr4_read_shadow;
	natural_width cr3_target_value0;
	natural_width cr3_target_value1;
	natural_width cr3_target_value2;
	natural_width cr3_target_value3;
	natural_width exit_qualification;
	natural_width guest_linear_address;
	natural_width guest_cr0;
	natural_width guest_cr3;
	natural_width guest_cr4;
	natural_width guest_es_base;
	natural_width guest_cs_base;
	natural_width guest_ss_base;
	natural_width guest_ds_base;
	natural_width guest_fs_base;
	natural_width guest_gs_base;
	natural_width guest_ldtr_base;
	natural_width guest_tr_base;
	natural_width guest_gdtr_base;
	natural_width guest_idtr_base;
	natural_width guest_dr7;
	natural_width guest_rsp;
	natural_width guest_rip;
	natural_width guest_rflags;
	natural_width guest_pending_dbg_exceptions;
	natural_width guest_sysenter_esp;
	natural_width guest_sysenter_eip;
	natural_width host_cr0;
	natural_width host_cr3;
	natural_width host_cr4;
	natural_width host_fs_base;
	natural_width host_gs_base;
	natural_width host_tr_base;
	natural_width host_gdtr_base;
	natural_width host_idtr_base;
	natural_width host_ia32_sysenter_esp;
	natural_width host_ia32_sysenter_eip;
	natural_width host_rsp;
	natural_width host_rip;
	natural_width paddingl[8]; /* room for future expansion */
	u32 pin_based_vm_exec_control;
	u32 cpu_based_vm_exec_control;
	u32 exception_bitmap;
	u32 page_fault_error_code_mask;
	u32 page_fault_error_code_match;
	u32 cr3_target_count;
	u32 vm_exit_controls;
	u32 vm_exit_msr_store_count;
	u32 vm_exit_msr_load_count;
	u32 vm_entry_controls;
	u32 vm_entry_msr_load_count;
	u32 vm_entry_intr_info_field;
	u32 vm_entry_exception_error_code;
	u32 vm_entry_instruction_len;
	u32 tpr_threshold;
	u32 secondary_vm_exec_control;
	u32 vm_instruction_error;
	u32 vm_exit_reason;
	u32 vm_exit_intr_info;
	u32 vm_exit_intr_error_code;
	u32 idt_vectoring_info_field;
	u32 idt_vectoring_error_code;
	u32 vm_exit_instruction_len;
	u32 vmx_instruction_info;
	u32 guest_es_limit;
	u32 guest_cs_limit;
	u32 guest_ss_limit;
	u32 guest_ds_limit;
	u32 guest_fs_limit;
	u32 guest_gs_limit;
	u32 guest_ldtr_limit;
	u32 guest_tr_limit;
	u32 guest_gdtr_limit;
	u32 guest_idtr_limit;
	u32 guest_es_ar_bytes;
	u32 guest_cs_ar_bytes;
	u32 guest_ss_ar_bytes;
	u32 guest_ds_ar_bytes;
	u32 guest_fs_ar_bytes;
	u32 guest_gs_ar_bytes;
	u32 guest_ldtr_ar_bytes;
	u32 guest_tr_ar_bytes;
	u32 guest_interruptibility_info;
	u32 guest_activity_state;
	u32 guest_sysenter_cs;
	u32 host_ia32_sysenter_cs;
	u32 vmx_preemption_timer_value;
	u32 padding32[7]; /* room for future expansion */
	u16 virtual_processor_id;
	u16 posted_intr_nv;
	u16 guest_es_selector;
	u16 guest_cs_selector;
	u16 guest_ss_selector;
	u16 guest_ds_selector;
	u16 guest_fs_selector;
	u16 guest_gs_selector;
	u16 guest_ldtr_selector;
	u16 guest_tr_selector;
	u16 guest_intr_status;
	u16 guest_pml_index;
	u16 host_es_selector;
	u16 host_cs_selector;
	u16 host_ss_selector;
	u16 host_ds_selector;
	u16 host_fs_selector;
	u16 host_gs_selector;
	u16 host_tr_selector;
};

struct pi_desc {
	u32 pir[8];     /* Posted interrupt requested */
	union {
		struct {
				/* bit 256 - Outstanding Notification */
			u16	on	: 1,
				/* bit 257 - Suppress Notification */
				sn	: 1,
				/* bit 271:258 - Reserved */
				rsvd_1	: 14;
				/* bit 279:272 - Notification Vector */
			u8	nv;
				/* bit 287:280 - Reserved */
			u8	rsvd_2;
				/* bit 319:288 - Notification Destination */
			u32	ndst;
		};
		u64 control;
	};
	u32 rsvd[6];
} __aligned(64);

struct vcpu_vmx {
	struct kvm_vcpu       vcpu;
	unsigned long         host_rsp;
	u8                    fail;
	u8		      msr_bitmap_mode;
	u32                   exit_intr_info;
	u32                   idt_vectoring_info;
	ulong                 rflags;
	struct shared_msr_entry *guest_msrs;
	int                   nmsrs;
	int                   save_nmsrs;
	unsigned long	      host_idt_base;
#ifdef CONFIG_X86_64
	u64 		      msr_host_kernel_gs_base;
	u64 		      msr_guest_kernel_gs_base;
#endif

	u64 		      arch_capabilities;
	u64 		      spec_ctrl;

	u32 vm_entry_controls_shadow;
	u32 vm_exit_controls_shadow;
	u32 secondary_exec_control;

	/*
	 * loaded_vmcs points to the VMCS currently used in this vcpu. For a
	 * non-nested (L1) guest, it always points to vmcs01. For a nested
	 * guest (L2), it points to a different VMCS.
	 */
	struct loaded_vmcs    vmcs01;
	struct loaded_vmcs   *loaded_vmcs;
	bool                  __launched; /* temporary, used in vmx_vcpu_run */
	struct msr_autoload {
		unsigned nr;
		struct vmx_msr_entry guest[KLP_NR_AUTOLOAD_MSRS];
		struct vmx_msr_entry host[KLP_NR_AUTOLOAD_MSRS];
	} msr_autoload;
	struct {
		int           loaded;
		u16           fs_sel, gs_sel, ldt_sel;
#ifdef CONFIG_X86_64
		u16           ds_sel, es_sel;
#endif
		int           gs_ldt_reload_needed;
		int           fs_reload_needed;
		u64           msr_host_bndcfgs;
	} host_state;
	struct {
		int vm86_active;
		ulong save_rflags;
		struct kvm_segment segs[8];
	} rmode;
	struct {
		u32 bitmask; /* 4 bits per segment (1 bit per field) */
		struct kvm_save_segment {
			u16 selector;
			unsigned long base;
			u32 limit;
			u32 ar;
		} seg[8];
	} segment_cache;
	int vpid;
	bool emulation_required;

	u32 exit_reason;

	/* Posted interrupt descriptor */
	struct pi_desc pi_desc;

	/* Support for a guest hypervisor (nested VMX) */
	struct nested_vmx nested;

	/* Dynamic PLE window. */
	int ple_window;
	bool ple_window_dirty;

	/* Support for PML */
#define PML_ENTITY_NUM		512
	struct page *pml_pg;

	/* apic deadline value in host tsc */
	u64 hv_deadline_tsc;

	u64 current_tsc_ratio;

	u32 host_pkru;

	/*
	 * Only bits masked by msr_ia32_feature_control_valid_bits can be set in
	 * msr_ia32_feature_control. FEATURE_CONTROL_LOCKED is always included
	 * in msr_ia32_feature_control_valid_bits.
	 */
	u64 msr_ia32_feature_control;
	u64 msr_ia32_feature_control_valid_bits;
};

/* inlined */
static inline struct vcpu_vmx *klp_to_vmx(struct kvm_vcpu *vcpu)
{
	return container_of(vcpu, struct vcpu_vmx, vcpu);
}

/* inlined */
static inline short klp_vmcs_field_to_offset(unsigned long field)
{
	const size_t size = ARRAY_SIZE(*klp_vmcs_field_to_offset_table);
	unsigned short offset;

	BUILD_BUG_ON(size > SHRT_MAX);
	if (field >= size)
		return -ENOENT;

	field = array_index_nospec(field, size);
	offset = (*klp_vmcs_field_to_offset_table)[field];
	if (offset == 0)
		return -ENOENT;
	return offset;
}

/* inlined */
static inline struct vmcs12 *klp_get_vmcs12(struct kvm_vcpu *vcpu)
{
	return klp_to_vmx(vcpu)->nested.cached_vmcs12;
}

/* inlined */
static inline bool klp_cpu_has_vmx_invept_context(void)
{
	return klp_vmx_capability->ept & VMX_EPT_EXTENT_CONTEXT_BIT;
}

/* inlined */
static inline bool klp_cpu_has_vmx_invvpid_single(void)
{
	return klp_vmx_capability->vpid & VMX_VPID_EXTENT_SINGLE_CONTEXT_BIT;
}

/* inlined */
static inline bool klp_cpu_has_vmx_invvpid_global(void)
{
	return klp_vmx_capability->vpid & VMX_VPID_EXTENT_GLOBAL_CONTEXT_BIT;
}

/* inlined */
static inline unsigned klp_nested_cpu_vmx_misc_cr3_count(struct kvm_vcpu *vcpu)
{
	return vmx_misc_cr3_count(klp_to_vmx(vcpu)->nested.nested_vmx_misc_low);
}

/* inlined */
static inline bool klp_nested_cpu_has(struct vmcs12 *vmcs12, u32 bit)
{
	return vmcs12->cpu_based_vm_exec_control & bit;
}

/* inlined */
static inline bool klp_nested_cpu_has2(struct vmcs12 *vmcs12, u32 bit)
{
	return (vmcs12->cpu_based_vm_exec_control &
			CPU_BASED_ACTIVATE_SECONDARY_CONTROLS) &&
		(vmcs12->secondary_vm_exec_control & bit);
}

/* inlined */
static inline int klp_nested_cpu_has_ept(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_EPT);
}

/* inlined */
static inline bool klp_nested_cpu_has_virt_x2apic_mode(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUALIZE_X2APIC_MODE);
}

/* inlined */
static inline bool klp_nested_cpu_has_apic_reg_virt(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_APIC_REGISTER_VIRT);
}

/* inlined */
static inline bool klp_nested_cpu_has_vid(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUAL_INTR_DELIVERY);
}

/* inlined */
static inline bool klp_nested_cpu_has_posted_intr(struct vmcs12 *vmcs12)
{
	return vmcs12->pin_based_vm_exec_control & PIN_BASED_POSTED_INTR;
}

/* inlined */
static inline bool klp_nested_cpu_has_vmfunc(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_VMFUNC);
}

/* inlined */
static inline bool klp_nested_cpu_has_eptp_switching(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has_vmfunc(vmcs12) &&
		(vmcs12->vm_function_control &
		 VMX_VMFUNC_EPTP_SWITCHING);
}

/* inlined */
static inline void klp__invvpid(int ext, u16 vpid, gva_t gva)
{
    struct {
	u64 vpid : 16;
	u64 rsvd : 48;
	u64 gva;
    } operand = { vpid, 0, gva };

    asm volatile (klp__ex(ASM_VMX_INVVPID)
		  /* CF==1 or ZF==1 --> rc = -1 */
		  "; ja 1f ; ud2 ; 1:"
		  : : [thunk_target] "a"(&operand), "c"(ext) : "cc", "memory");
}

/* inlined */
static inline void klp__invept(int ext, u64 eptp, gpa_t gpa)
{
	struct {
		u64 eptp, gpa;
	} operand = {eptp, gpa};

	asm volatile (klp__ex(ASM_VMX_INVEPT)
			/* CF==1 or ZF==1 --> rc = -1 */
			"; ja 1f ; ud2 ; 1:\n"
			: : [thunk_target] "a" (&operand), "c" (ext) : "cc", "memory");
}

/* inlined */
static inline void klp_vpid_sync_vcpu_single(int vpid)
{
	if (vpid == 0)
		return;

	if (klp_cpu_has_vmx_invvpid_single())
		klp__invvpid(VMX_VPID_EXTENT_SINGLE_CONTEXT, vpid, 0);
}

/* inlined */
static inline void klp_vpid_sync_vcpu_global(void)
{
	if (klp_cpu_has_vmx_invvpid_global())
		klp__invvpid(VMX_VPID_EXTENT_ALL_CONTEXT, 0, 0);
}

/* inlined */
static inline void klp_vpid_sync_context(int vpid)
{
	if (klp_cpu_has_vmx_invvpid_single())
		klp_vpid_sync_vcpu_single(vpid);
	else
		klp_vpid_sync_vcpu_global();
}

/* inlined */
static inline void klp_ept_sync_global(void)
{
	klp__invept(VMX_EPT_EXTENT_GLOBAL, 0, 0);
}

/* inlined */
static inline void klp_ept_sync_context(u64 eptp)
{
	if (klp_cpu_has_vmx_invept_context())
		klp__invept(VMX_EPT_EXTENT_CONTEXT, eptp, 0);
	else
		klp_ept_sync_global();
}

/* inlined */
static __always_inline void klp_vmcs_check16(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2000,
			 "16-bit accessor invalid for 64-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2001,
			 "16-bit accessor invalid for 64-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x4000,
			 "16-bit accessor invalid for 32-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x6000,
			 "16-bit accessor invalid for natural width field");
}

/* inlined */
static __always_inline void klp_vmcs_check32(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0,
			 "32-bit accessor invalid for 16-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x6000,
			 "32-bit accessor invalid for natural width field");
}

/* inlined */
static __always_inline void klp_vmcs_check64(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0,
			 "64-bit accessor invalid for 16-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2001,
			 "64-bit accessor invalid for 64-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x4000,
			 "64-bit accessor invalid for 32-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x6000,
			 "64-bit accessor invalid for natural width field");
}

/* inlined */
static __always_inline void klp_vmcs_checkl(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0,
			 "Natural width accessor invalid for 16-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2000,
			 "Natural width accessor invalid for 64-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2001,
			 "Natural width accessor invalid for 64-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x4000,
			 "Natural width accessor invalid for 32-bit field");
}

/* inlined */
static __always_inline unsigned long klp__vmcs_readl(unsigned long field)
{
	unsigned long value;

	asm volatile (klp__ex_clear(ASM_VMX_VMREAD_RDX_RAX, "%0")
		      : [thunk_target] "=a"(value) : "d"(field) : "cc");
	return value;
}

/* inlined */
static __always_inline void klp__vmcs_writel(unsigned long field,
					     unsigned long value)
{
	u8 error;

	asm volatile (klp__ex(ASM_VMX_VMWRITE_RAX_RDX) "; setna %0"
		       : "=q"(error) : [thunk_target] "a"(value), "d"(field) : "cc");
	if (unlikely(error))
		klp_vmwrite_error(field, value);
}

/* inlined */
static __always_inline void klp_vmcs_write64(unsigned long field, u64 value)
{
	klp_vmcs_check64(field);
	klp__vmcs_writel(field, value);
#ifndef CONFIG_X86_64
	asm volatile ("");
	klp__vmcs_writel(field+1, value >> 32);
#endif
}

/* inlined */
static __always_inline void klp_vmcs_clear_bits(unsigned long field, u32 mask)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x2000,
			 "vmcs_clear_bits does not support 64-bit fields");
	klp__vmcs_writel(field, klp__vmcs_readl(field) & ~mask);
}

/* inlined */
static __always_inline void klp_vmcs_set_bits(unsigned long field, u32 mask)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x2000,
			 "vmcs_set_bits does not support 64-bit fields");
	klp__vmcs_writel(field, klp__vmcs_readl(field) | mask);
}

/* inlined */
static __always_inline u32 klp_vmcs_read32(unsigned long field)
{
	klp_vmcs_check32(field);
	return klp__vmcs_readl(field);
}

/* inlined */
static __always_inline unsigned long klp_vmcs_readl(unsigned long field)
{
	klp_vmcs_checkl(field);
	return klp__vmcs_readl(field);
}

/* inlined */
static inline bool klp_fixed_bits_valid(u64 val, u64 fixed0, u64 fixed1)
{
	return ((val & fixed1) | fixed0) == val;
}

/* inlined */
static inline bool klp_vmx_control_verify(u32 control, u32 low, u32 high)
{
	return klp_fixed_bits_valid(control, low, high);
}

/* inlined */
static struct vmcs *klp_alloc_vmcs(void)
{
	return klp_alloc_vmcs_cpu(raw_smp_processor_id());
}

enum vmcs_field_type {
	VMCS_FIELD_TYPE_U16 = 0,
	VMCS_FIELD_TYPE_U64 = 1,
	VMCS_FIELD_TYPE_U32 = 2,
	VMCS_FIELD_TYPE_NATURAL_WIDTH = 3
};

/* inlined */
static inline int klp_vmcs_field_type(unsigned long field)
{
	if (0x1 & field)	/* the *_HIGH fields are all 32 bit */
		return VMCS_FIELD_TYPE_U32;
	return (field >> 13) & 0x3 ;
}

/* inlined */
static inline int klp_vmcs_field_readonly(unsigned long field)
{
	return (((field >> 10) & 0x3) == 1);
}

/* inlined */
static inline void klp__vmx_flush_tlb(struct kvm_vcpu *vcpu, int vpid)
{
	if (*klp_enable_ept) {
		if (!VALID_PAGE(vcpu->arch.mmu.root_hpa))
			return;
		klp_ept_sync_context(klp_construct_eptp(vcpu, vcpu->arch.mmu.root_hpa));
	} else {
		klp_vpid_sync_context(vpid);
	}
}

/* inlined */
static int klp_enter_vmx_operation(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	struct vmcs *shadow_vmcs;
	int r;

	r = klp_alloc_loaded_vmcs(&vmx->nested.vmcs02);
	if (r < 0)
		goto out_vmcs02;

	vmx->nested.cached_vmcs12 = kmalloc(KLP_VMCS12_SIZE, GFP_KERNEL);
	if (!vmx->nested.cached_vmcs12)
		goto out_cached_vmcs12;

	if (*klp_enable_shadow_vmcs) {
		shadow_vmcs = klp_alloc_vmcs();
		if (!shadow_vmcs)
			goto out_shadow_vmcs;
		/* mark vmcs as shadow */
		shadow_vmcs->revision_id |= (1u << 31);
		/* init shadow vmcs */
		klp_vmcs_clear(shadow_vmcs);
		vmx->vmcs01.shadow_vmcs = shadow_vmcs;
	}

	hrtimer_init(&vmx->nested.preemption_timer, CLOCK_MONOTONIC,
		     HRTIMER_MODE_REL_PINNED);
	vmx->nested.preemption_timer.function = klp_vmx_preemption_timer_fn;

	vmx->nested.vmxon = true;
	return 0;

out_shadow_vmcs:
	kfree(vmx->nested.cached_vmcs12);

out_cached_vmcs12:
	klp_free_loaded_vmcs(&vmx->nested.vmcs02);

out_vmcs02:
	return -ENOMEM;
}

/* optimized */
static void klp_vmx_disable_shadow_vmcs(struct vcpu_vmx *vmx)
{
	klp_vmcs_clear_bits(SECONDARY_VM_EXEC_CONTROL, SECONDARY_EXEC_SHADOW_VMCS);
	klp_vmcs_write64(VMCS_LINK_POINTER, -1ull);
}

/* inlined */
static inline void klp_nested_release_vmcs12(struct vcpu_vmx *vmx)
{
	if (vmx->nested.current_vmptr == -1ull)
		return;

	if (*klp_enable_shadow_vmcs) {
		/* copy to memory all shadowed fields in case
		   they were modified */
		klp_copy_shadow_to_vmcs12(vmx);
		vmx->nested.sync_shadow_vmcs = false;
		klp_vmx_disable_shadow_vmcs(vmx);
	}
	vmx->nested.posted_intr_nv = -1;

	/* Flush VMCS12 to guest memory */
	klp_kvm_vcpu_write_guest_page(&vmx->vcpu,
				     vmx->nested.current_vmptr >> PAGE_SHIFT,
				     vmx->nested.cached_vmcs12, 0, KLP_VMCS12_SIZE);

	vmx->nested.current_vmptr = -1ull;
}

/* inlined */
static inline int klp_vmcs12_read_any(struct kvm_vcpu *vcpu,
				      unsigned long field, u64 *ret)
{
	short offset = klp_vmcs_field_to_offset(field);
	char *p;

	if (offset < 0)
		return offset;

	p = ((char *)(klp_get_vmcs12(vcpu))) + offset;

	switch (klp_vmcs_field_type(field)) {
	case VMCS_FIELD_TYPE_NATURAL_WIDTH:
		*ret = *((natural_width *)p);
		return 0;
	case VMCS_FIELD_TYPE_U16:
		*ret = *((u16 *)p);
		return 0;
	case VMCS_FIELD_TYPE_U32:
		*ret = *((u32 *)p);
		return 0;
	case VMCS_FIELD_TYPE_U64:
		*ret = *((u64 *)p);
		return 0;
	default:
		WARN_ON(1);
		return -ENOENT;
	}
}

/* inlined */
static inline int klp_vmcs12_write_any(struct kvm_vcpu *vcpu,
				       unsigned long field, u64 field_value){
	short offset = klp_vmcs_field_to_offset(field);
	char *p = ((char *) klp_get_vmcs12(vcpu)) + offset;
	if (offset < 0)
		return offset;

	switch (klp_vmcs_field_type(field)) {
	case VMCS_FIELD_TYPE_U16:
		*(u16 *)p = field_value;
		return 0;
	case VMCS_FIELD_TYPE_U32:
		*(u32 *)p = field_value;
		return 0;
	case VMCS_FIELD_TYPE_U64:
		*(u64 *)p = field_value;
		return 0;
	case VMCS_FIELD_TYPE_NATURAL_WIDTH:
		*(natural_width *)p = field_value;
		return 0;
	default:
		WARN_ON(1);
		return -ENOENT;
	}

}

/* inlined */
static int klp_nested_vmx_check_vmcs12(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	if (vmx->nested.current_vmptr == -1ull) {
		klp_nested_vmx_failInvalid(vcpu);
		return 0;
	}
	return 1;
}

/* inlined */
static void klp_set_current_vmptr(struct vcpu_vmx *vmx, gpa_t vmptr)
{
	vmx->nested.current_vmptr = vmptr;
	if (*klp_enable_shadow_vmcs) {
		klp_vmcs_set_bits(SECONDARY_VM_EXEC_CONTROL,
				  SECONDARY_EXEC_SHADOW_VMCS);
		klp_vmcs_write64(VMCS_LINK_POINTER,
				 __pa(vmx->vmcs01.shadow_vmcs));
		vmx->nested.sync_shadow_vmcs = true;
	}
}



/* patched */
int klp_handle_vmon(struct kvm_vcpu *vcpu)
{
	int ret;
	gpa_t vmptr;
	struct page *page;
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	const u64 VMXON_NEEDED_FEATURES = FEATURE_CONTROL_LOCKED
		| FEATURE_CONTROL_VMXON_ENABLED_OUTSIDE_SMX;

	/*
	 * The Intel VMX Instruction Reference lists a bunch of bits that are
	 * prerequisite to running VMXON, most notably cr4.VMXE must be set to
	 * 1 (see vmx_set_cr4() for when we allow the guest to set this).
	 * Otherwise, we should fail with #UD.  But most faulting conditions
	 * have already been checked by hardware, prior to the VM-exit for
	 * VMXON.  We do test guest cr4.VMXE because processor CR4 always has
	 * that bit set to 1 in non-root mode.
	 */
	if (!klp_kvm_read_cr4_bits(vcpu, X86_CR4_VMXE)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 1;
	}

	/*
	 * Fix CVE-2018-12904
	 *  +6 lines
	 */
	/* CPL=0 must be checked manually. */
	if (klp_vmx_get_cpl(vcpu)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 1;
	}

	if (vmx->nested.vmxon) {
		klp_nested_vmx_failValid(vcpu, VMXERR_VMXON_IN_VMX_ROOT_OPERATION);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if ((vmx->msr_ia32_feature_control & VMXON_NEEDED_FEATURES)
			!= VMXON_NEEDED_FEATURES) {
		klp_kvm_inject_gp(vcpu, 0);
		return 1;
	}

	if (klp_nested_vmx_get_vmptr(vcpu, &vmptr))
		return 1;

	/*
	 * SDM 3: 24.11.5
	 * The first 4 bytes of VMXON region contain the supported
	 * VMCS revision identifier
	 *
	 * Note - IA32_VMX_BASIC[48] will never be 1 for the nested case;
	 * which replaces physical address width with 32
	 */
	if (!PAGE_ALIGNED(vmptr) || (vmptr >> klp_cpuid_maxphyaddr(vcpu))) {
		klp_nested_vmx_failInvalid(vcpu);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	page = klp_kvm_vcpu_gpa_to_page(vcpu, vmptr);
	if (is_error_page(page)) {
		klp_nested_vmx_failInvalid(vcpu);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}
	if (*(u32 *)kmap(page) != KLP_VMCS12_REVISION) {
		kunmap(page);
		klp_kvm_release_page_clean(page);
		klp_nested_vmx_failInvalid(vcpu);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}
	kunmap(page);
	klp_kvm_release_page_clean(page);

	vmx->nested.vmxon_ptr = vmptr;
	ret = klp_enter_vmx_operation(vcpu);
	if (ret)
		return ret;

	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, inlined */
static int klp_nested_vmx_check_permission(struct kvm_vcpu *vcpu)
{
	/*
	 * Fix CVE-2018-12904
	 *  +5 lines
	 */
	if (klp_vmx_get_cpl(vcpu)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 0;
	}

	if (!klp_to_vmx(vcpu)->nested.vmxon) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 0;
	}
	return 1;
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmoff(struct kvm_vcpu *vcpu)
{
	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;
	klp_free_nested(klp_to_vmx(vcpu));
	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmclear(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	u32 zero = 0;
	gpa_t vmptr;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (klp_nested_vmx_get_vmptr(vcpu, &vmptr))
		return 1;

	if (!PAGE_ALIGNED(vmptr) || (vmptr >> klp_cpuid_maxphyaddr(vcpu))) {
		klp_nested_vmx_failValid(vcpu, VMXERR_VMCLEAR_INVALID_ADDRESS);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if (vmptr == vmx->nested.vmxon_ptr) {
		klp_nested_vmx_failValid(vcpu, VMXERR_VMCLEAR_VMXON_POINTER);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if (vmptr == vmx->nested.current_vmptr)
		klp_nested_release_vmcs12(vmx);

	klp_kvm_vcpu_write_guest(vcpu,
			vmptr + offsetof(struct vmcs12, launch_state),
			&zero, sizeof(zero));

	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmread(struct kvm_vcpu *vcpu)
{
	unsigned long field;
	u64 field_value;
	unsigned long exit_qualification = klp_vmcs_readl(EXIT_QUALIFICATION);
	u32 vmx_instruction_info = klp_vmcs_read32(VMX_INSTRUCTION_INFO);
	gva_t gva = 0;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (!klp_nested_vmx_check_vmcs12(vcpu))
		return klp_kvm_skip_emulated_instruction(vcpu);

	/* Decode instruction info and find the field to read */
	field = klp_kvm_register_readl(vcpu, (((vmx_instruction_info) >> 28) & 0xf));
	/* Read the field, zero-extended to a u64 field_value */
	if (klp_vmcs12_read_any(vcpu, field, &field_value) < 0) {
		klp_nested_vmx_failValid(vcpu, VMXERR_UNSUPPORTED_VMCS_COMPONENT);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}
	/*
	 * Now copy part of this value to register or memory, as requested.
	 * Note that the number of bits actually copied is 32 or 64 depending
	 * on the guest's mode (32 or 64 bit), not on the given field's length.
	 */
	if (vmx_instruction_info & (1u << 10)) {
		klp_kvm_register_writel(vcpu, (((vmx_instruction_info) >> 3) & 0xf),
			field_value);
	} else {
		if (klp_get_vmx_mem_address(vcpu, exit_qualification,
				vmx_instruction_info, true, &gva))
			return 1;
		/* _system ok, as hardware has verified cpl=0 */
		klp_kvm_write_guest_virt_system(&vcpu->arch.emulate_ctxt, gva,
			     &field_value, (klp_is_long_mode(vcpu) ? 8 : 4), NULL);
	}

	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmwrite(struct kvm_vcpu *vcpu)
{
	unsigned long field;
	gva_t gva;
	unsigned long exit_qualification = klp_vmcs_readl(EXIT_QUALIFICATION);
	u32 vmx_instruction_info = klp_vmcs_read32(VMX_INSTRUCTION_INFO);
	/* The value to write might be 32 or 64 bits, depending on L1's long
	 * mode, and eventually we need to write that into a field of several
	 * possible lengths. The code below first zero-extends the value to 64
	 * bit (field_value), and then copies only the appropriate number of
	 * bits into the vmcs12 field.
	 */
	u64 field_value = 0;
	struct x86_exception e;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (!klp_nested_vmx_check_vmcs12(vcpu))
		return klp_kvm_skip_emulated_instruction(vcpu);

	if (vmx_instruction_info & (1u << 10))
		field_value = klp_kvm_register_readl(vcpu,
			(((vmx_instruction_info) >> 3) & 0xf));
	else {
		if (klp_get_vmx_mem_address(vcpu, exit_qualification,
				vmx_instruction_info, false, &gva))
			return 1;
		if (klp_kvm_read_guest_virt(&vcpu->arch.emulate_ctxt, gva,
			   &field_value, (klp_is_64_bit_mode(vcpu) ? 8 : 4), &e)) {
			klp_kvm_inject_page_fault(vcpu, &e);
			return 1;
		}
	}


	field = klp_kvm_register_readl(vcpu, (((vmx_instruction_info) >> 28) & 0xf));
	if (klp_vmcs_field_readonly(field)) {
		klp_nested_vmx_failValid(vcpu,
			VMXERR_VMWRITE_READ_ONLY_VMCS_COMPONENT);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if (klp_vmcs12_write_any(vcpu, field, field_value) < 0) {
		klp_nested_vmx_failValid(vcpu, VMXERR_UNSUPPORTED_VMCS_COMPONENT);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmptrld(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	gpa_t vmptr;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (klp_nested_vmx_get_vmptr(vcpu, &vmptr))
		return 1;

	if (!PAGE_ALIGNED(vmptr) || (vmptr >> klp_cpuid_maxphyaddr(vcpu))) {
		klp_nested_vmx_failValid(vcpu, VMXERR_VMPTRLD_INVALID_ADDRESS);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if (vmptr == vmx->nested.vmxon_ptr) {
		klp_nested_vmx_failValid(vcpu, VMXERR_VMPTRLD_VMXON_POINTER);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	if (vmx->nested.current_vmptr != vmptr) {
		struct vmcs12 *new_vmcs12;
		struct page *page;
		page = klp_kvm_vcpu_gpa_to_page(vcpu, vmptr);
		if (is_error_page(page)) {
			klp_nested_vmx_failInvalid(vcpu);
			return klp_kvm_skip_emulated_instruction(vcpu);
		}
		new_vmcs12 = kmap(page);
		if (new_vmcs12->revision_id != KLP_VMCS12_REVISION) {
			kunmap(page);
			klp_kvm_release_page_clean(page);
			klp_nested_vmx_failValid(vcpu,
				VMXERR_VMPTRLD_INCORRECT_VMCS_REVISION_ID);
			return klp_kvm_skip_emulated_instruction(vcpu);
		}

		klp_nested_release_vmcs12(vmx);
		/*
		 * Load VMCS12 from guest memory since it is not already
		 * cached.
		 */
		memcpy(vmx->nested.cached_vmcs12, new_vmcs12, KLP_VMCS12_SIZE);
		kunmap(page);
		klp_kvm_release_page_clean(page);

		klp_set_current_vmptr(vmx, vmptr);
	}

	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_vmptrst(struct kvm_vcpu *vcpu)
{
	unsigned long exit_qualification = klp_vmcs_readl(EXIT_QUALIFICATION);
	u32 vmx_instruction_info = klp_vmcs_read32(VMX_INSTRUCTION_INFO);
	gva_t vmcs_gva;
	struct x86_exception e;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (klp_get_vmx_mem_address(vcpu, exit_qualification,
			vmx_instruction_info, true, &vmcs_gva))
		return 1;
	/* ok to use *_system, as hardware has verified cpl=0 */
	if (klp_kvm_write_guest_virt_system(&vcpu->arch.emulate_ctxt, vmcs_gva,
				 (void *)&klp_to_vmx(vcpu)->nested.current_vmptr,
				 sizeof(u64), &e)) {
		klp_kvm_inject_page_fault(vcpu, &e);
		return 1;
	}
	klp_nested_vmx_succeed(vcpu);
	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_invept(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	u32 vmx_instruction_info, types;
	unsigned long type;
	gva_t gva;
	struct x86_exception e;
	struct {
		u64 eptp, gpa;
	} operand;

	if (!(vmx->nested.nested_vmx_secondary_ctls_high &
	      SECONDARY_EXEC_ENABLE_EPT) ||
	    !(vmx->nested.nested_vmx_ept_caps & VMX_EPT_INVEPT_BIT)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 1;
	}

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	vmx_instruction_info = klp_vmcs_read32(VMX_INSTRUCTION_INFO);
	type = klp_kvm_register_readl(vcpu, (vmx_instruction_info >> 28) & 0xf);

	types = (vmx->nested.nested_vmx_ept_caps >> VMX_EPT_EXTENT_SHIFT) & 6;

	if (type >= 32 || !(types & (1 << type))) {
		klp_nested_vmx_failValid(vcpu,
				VMXERR_INVALID_OPERAND_TO_INVEPT_INVVPID);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	/* According to the Intel VMX instruction reference, the memory
	 * operand is read even if it isn't needed (e.g., for type==global)
	 */
	if (klp_get_vmx_mem_address(vcpu, klp_vmcs_readl(EXIT_QUALIFICATION),
			vmx_instruction_info, false, &gva))
		return 1;
	if (klp_kvm_read_guest_virt(&vcpu->arch.emulate_ctxt, gva, &operand,
				    sizeof(operand), &e)) {
		klp_kvm_inject_page_fault(vcpu, &e);
		return 1;
	}

	switch (type) {
	case VMX_EPT_EXTENT_GLOBAL:
	/*
	 * TODO: track mappings and invalidate
	 * single context requests appropriately
	 */
	case VMX_EPT_EXTENT_CONTEXT:
		klp_kvm_mmu_sync_roots(vcpu);
		kvm_make_request(KVM_REQ_TLB_FLUSH, vcpu);
		klp_nested_vmx_succeed(vcpu);
		break;
	default:
		BUG_ON(1);
		break;
	}

	return klp_kvm_skip_emulated_instruction(vcpu);
}

/* patched, calls nested_vmx_check_permission() */
int klp_handle_invvpid(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	u32 vmx_instruction_info;
	unsigned long type, types;
	gva_t gva;
	struct x86_exception e;
	struct {
		u64 vpid;
		u64 gla;
	} operand;

	if (!(vmx->nested.nested_vmx_secondary_ctls_high &
	      SECONDARY_EXEC_ENABLE_VPID) ||
			!(vmx->nested.nested_vmx_vpid_caps & VMX_VPID_INVVPID_BIT)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 1;
	}

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	vmx_instruction_info = klp_vmcs_read32(VMX_INSTRUCTION_INFO);
	type = klp_kvm_register_readl(vcpu, (vmx_instruction_info >> 28) & 0xf);

	types = (vmx->nested.nested_vmx_vpid_caps &
			KLP_VMX_VPID_EXTENT_SUPPORTED_MASK) >> 8;

	if (type >= 32 || !(types & (1 << type))) {
		klp_nested_vmx_failValid(vcpu,
			VMXERR_INVALID_OPERAND_TO_INVEPT_INVVPID);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	/* according to the intel vmx instruction reference, the memory
	 * operand is read even if it isn't needed (e.g., for type==global)
	 */
	if (klp_get_vmx_mem_address(vcpu, klp_vmcs_readl(EXIT_QUALIFICATION),
			vmx_instruction_info, false, &gva))
		return 1;
	if (klp_kvm_read_guest_virt(&vcpu->arch.emulate_ctxt, gva, &operand,
				    sizeof(operand), &e)) {
		klp_kvm_inject_page_fault(vcpu, &e);
		return 1;
	}
	if (operand.vpid >> 16) {
		klp_nested_vmx_failValid(vcpu,
			VMXERR_INVALID_OPERAND_TO_INVEPT_INVVPID);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	switch (type) {
	case VMX_VPID_EXTENT_INDIVIDUAL_ADDR:
		if (klp_is_noncanonical_address(operand.gla, vcpu)) {
			klp_nested_vmx_failValid(vcpu,
				VMXERR_INVALID_OPERAND_TO_INVEPT_INVVPID);
			return klp_kvm_skip_emulated_instruction(vcpu);
		}
		/* fall through */
	case VMX_VPID_EXTENT_SINGLE_CONTEXT:
	case VMX_VPID_EXTENT_SINGLE_NON_GLOBAL:
		if (!operand.vpid) {
			klp_nested_vmx_failValid(vcpu,
				VMXERR_INVALID_OPERAND_TO_INVEPT_INVVPID);
			return klp_kvm_skip_emulated_instruction(vcpu);
		}
		break;
	case VMX_VPID_EXTENT_ALL_CONTEXT:
		break;
	default:
		WARN_ON_ONCE(1);
		return klp_kvm_skip_emulated_instruction(vcpu);
	}

	klp__vmx_flush_tlb(vcpu, vmx->nested.vpid02);
	klp_nested_vmx_succeed(vcpu);

	return klp_kvm_skip_emulated_instruction(vcpu);
}



static int livepatch_bsc1099258_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(klp_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(klp_funcs[i].name);
		if (!addr) {
			pr_err("livepatch: symbol %s not resolved\n",
				klp_funcs[i].name);
			return -ENOENT;
		}

		*(klp_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int livepatch_bsc1099258_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KLP_PATCHED_MODULE))
		return 0;

	ret = livepatch_bsc1099258_kallsyms();
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1099258_module_nb = {
	.notifier_call = livepatch_bsc1099258_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1099258_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KLP_PATCHED_MODULE)) {
		ret = livepatch_bsc1099258_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&livepatch_bsc1099258_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1099258_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1099258_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM) */
